# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for job operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the job entity. The tests ensure that the queues can be
registered, renamed, queried, and deleted as expected through the REST API.
"""
from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from pytest import MonkeyPatch
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_EXPERIMENTS_ROUTE, V1_JOBS_ROUTE, V1_ROOT

from ..lib import actions, asserts, helpers, mock_rq

# -- Actions ---------------------------------------------------------------------------


def delete_job(
    client: FlaskClient,
    job_id: int,
) -> TestResponse:
    """Delete a job using the API.

    Args:
        client: The Flask test client.
        job_id: The id of the job to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}/{job_id}",
        follow_redirects=True,
    )


def modify_job_status(
    client: FlaskClient,
    experiment_id: int,
    job_id: int,
    new_status: str,
) -> TestResponse:
    """Change the status of a job using the API.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment the job belongs too.
        job_id: The id of the job to delete.
        new_status: The new status of the job.

    Returns:
        The response from the API.
    """
    payload = {"status": new_status}
    return client.put(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}/jobs/{job_id}/status",
        json=payload,
        follow_redirects=True,
    )


def get_job(
    client: FlaskClient,
    job_id: int,
) -> TestResponse:
    """Get a job using the API.

    Args:
        client: The Flask test client.
        job_id: The id of the job to get.

    Returns:
        The response from the API.
    """
    return client.get(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}/{job_id}",
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_job_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that a job response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    # Check expected keys
    expected_keys = {
        "id",
        "snapshot",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "description",
        "timeout",
        "status",
        "values",
        "user",
        "group",
        "tags",
        "queue",
        "experiment",
        "entrypoint",
        "artifacts",
    }
    assert set(response.keys()) == expected_keys

    # Check basic response types for base resources
    asserts.assert_base_resource_contents_match_expectations(response)

    # Check basic response types for job
    assert isinstance(response["description"], str)
    assert isinstance(response["timeout"], str)
    assert isinstance(response["values"], dict)

    assert response["description"] == expected_contents["description"]
    assert response["timeout"] == expected_contents["timeout"]
    assert response["values"] == expected_contents["values"]

    assert helpers.is_timeout_format(response["timeout"])

    # Check Refs for base resources
    asserts.assert_user_ref_contents_matches_expectations(
        user=response["user"], expected_user_id=expected_contents["user_id"]
    )
    asserts.assert_group_ref_contents_matches_expectations(
        group=response["group"], expected_group_id=expected_contents["group_id"]
    )
    asserts.assert_tag_ref_contents_matches_expectations(tags=response["tags"])

    # Check Refs for jobs
    asserts.assert_queue_ref_contents_matches_expectations(
        queue=response["queue"],
        expected_queue_id=expected_contents["queue_id"],
        expected_group_id=expected_contents["group_id"],
    )
    asserts.assert_experiment_ref_contents_matches_expectations(
        experiment=response["experiment"],
        expected_experiment_id=expected_contents["experiment_id"],
        expected_group_id=expected_contents["group_id"],
    )
    asserts.assert_entrypoint_ref_contents_matches_expectations(
        entrypoint=response["entrypoint"],
        expected_entrypoint_id=expected_contents["entrypoint_id"],
        expected_group_id=expected_contents["group_id"],
    )


def assert_retrieving_job_by_id_works(
    client: FlaskClient, job_id: int, expected: dict[str, Any]
) -> None:
    """Assert that retrieving a job by id works.

    Args:
        client: The Flask test client.
        job_id: The id of the job to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(f"/{V1_ROOT}/{V1_JOBS_ROUTE}/{job_id}", follow_redirects=True)
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_jobs_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all jobs works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    query_string: dict[str, Any] = {}

    if group_id:
        query_string["groupId"] = group_id

    if search:
        query_string["search"] = search

    if paging_info:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_job_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that jobs can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of job ids after sorting.
            See test_job_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    job_ids = [job["id"] for job in response_data["data"]]

    assert response.status_code == 200 and job_ids == expected


def assert_job_is_not_found(
    client: FlaskClient,
    job_id: int,
) -> None:
    """Assert that a job is not found.

    Args:
        client: The Flask test client.
        job_id: The id of the job to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}/{job_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_job_status_matches_expectations(
    client: FlaskClient, job_id: int, expected: str
) -> None:
    response = client.get(
        f"/{V1_ROOT}/{V1_JOBS_ROUTE}/{job_id}/status",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["status"] == expected


# -- Tests -----------------------------------------------------------------------------


def test_create_job(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_queues: dict[str, Any],
    registered_experiments: dict[str, Any],
    registered_entrypoints: dict[str, Any],
    monkeypatch: MonkeyPatch,
) -> None:
    """Test that jobs can be correctly registered and retrieved using the API.

    Given an authenticated user, registered queues, registered experiments, and
    registered entrypoints, this test validates the following sequence of actions:

    - The user registers a job.
    - The response is valid and matches the expected values given the registration
      request.
    - The user is able to retrieve information about the job using the job id.
    """

    """
    Register a Job
    ==============

    actions.py::register_job(*args: JobSchema)):
      JobSchema:
        From base resource schema:
          - group_id: The ID of the group the job belongs to.
        From job schema:
          - description: The description of the job
          - queue_id: The ID of queue the job is to run on.
          - experiment_id: The ID of the experiment the job belongs to.
          - entrypoint_id: The ID of the entrypoint that the job calls.
          - values: The values the job supplies to the entrypoint
          - timeout: The timeout value for the job to terminate if not completed by.
    """
    # Inline import necessary to prevent circular import
    import dioptra.restapi.v1.shared.rq_service as rq_service

    monkeypatch.setattr(rq_service, "RQQueue", mock_rq.MockRQQueue)

    description = "The new job."
    queue_id = registered_queues["queue1"]["snapshot"]
    experiment_id = registered_experiments["experiment1"]["snapshot"]
    entrypoint_id = registered_entrypoints["entrypoint1"]["snapshot"]
    values = {
        registered_entrypoints["entrypoint1"]["parameters"][0]["name"]: "new_value",
    }
    timeout = "24h"

    job_response = actions.register_job(
        client=client,
        queue_id=queue_id,
        experiment_id=experiment_id,
        entrypoint_id=entrypoint_id,
        description=description,
        values=values,
        timeout=timeout,
    ).get_json()

    """
    Validate the response matches the expected contents
    ===================================================

    response: The response from actions.py::register_job()
    expected_contents: The raw data passed to actions.py::register_job() as *args
      *Note: group_id is given as an arg for registration in the service layer
    """
    for param in registered_entrypoints["entrypoint1"]["parameters"][1:]:
        values[param["name"]] = param["defaultValue"]
    assert_job_response_contents_matches_expectations(
        response=job_response,
        expected_contents={
            "description": description,
            "timeout": timeout,
            "values": values,
            "user_id": auth_account["id"],
            "group_id": registered_experiments["experiment1"]["group"]["id"],
            "queue_id": queue_id,
            "experiment_id": experiment_id,
            "entrypoint_id": entrypoint_id,
        },
    )

    """
    Validate the job can be retrieved from the database
    ===================================================

    job_id: The id of the job
    expected: The response from the calling get on the database should match
      the response of the actions.py::register_job()
    """
    assert_retrieving_job_by_id_works(
        client, job_id=job_response["id"], expected=job_response
    )


def test_job_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that all jobs can be retrieved.

    Given an authenticated user and registered jobs this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered jobs.
    - The returned list of jobs matches the full list of registered jobs.
    """
    job_expected_list = list(registered_jobs.values())
    assert_retrieving_jobs_works(client, expected=job_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["job1", "job2", "job3"]),
        ("description", True, ["job2", "job1", "job3"]),
        ("description", False, ["job3", "job1", "job2"]),
        ("createdOn", True, ["job3", "job2", "job1"]),
        ("createdOn", False, ["job1", "job2", "job3"]),
    ],
)
def test_job_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that jobs can be sorted by column.

    Given an authenticated user and registered jobs, this test validates the following
    sequence of actions:

    - A user registers three jobs descriptions:
      "The first job.",
      "The second job.",
      "Not retrieved.".
    - The user is able to retrieve a list of all registered jobs sorted by a column
      ascending/descending.
    - The returned list of jobs matches the order in the parametrize lists above.
    """

    expected_ids = [registered_jobs[expected_name]["id"] for expected_name in expected]
    assert_sorting_job_works(client, sortBy, descending, expected=expected_ids)


def test_job_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that jobs can be queried with a search term.

    Given an authenticated user and registered jobs this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered jobs with a query on its
      description.
    - The returned list of jobs matches the expected list from the query.
    """
    job_expected_list = list(registered_jobs.values())[:2]
    assert_retrieving_jobs_works(
        client, expected=job_expected_list, search="description:*job*"
    )
    jobs_expected_list = list(registered_jobs.values())
    assert_retrieving_jobs_works(client, expected=jobs_expected_list, search="*")


def test_job_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that jobs can be queried with a group filter.

    Given an authenticated user and registered jobs this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered jobs with the specified
      group id.
    - The returned list of jobs matches the expected list from the query.
    """
    job_expected_list = list(registered_jobs.values())
    assert_retrieving_jobs_works(
        client,
        expected=job_expected_list,
        group_id=auth_account["groups"][0]["id"],
    )


def test_delete_job(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that a job can be deleted by referencing its id.

    Given an authenticated user and registered jobs, this test validates the following
    sequence of actions:

    - The user deletes a job by referencing its id.
    - The user attempts to retrieve information about the deleted job.
    - The request fails with an appropriate error message and response code.
    """
    job_to_delete = registered_jobs["job1"]
    delete_job(client, job_id=job_to_delete["id"])
    assert_job_is_not_found(client, job_id=job_to_delete["id"])


def test_job_get_status(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that the status of a job can be retrieved by the user.

    Given an authenticated user and registered jobs, this test validates the following
    sequence of actions:

    - The user is able to retrieve the status of a registered job.
    - The returned job status matches the expected status.
    """
    job_to_check = registered_jobs["job1"]
    status = "queued"
    assert_job_status_matches_expectations(
        client, job_id=job_to_check["id"], expected=status
    )


def test_modify_job_status(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that the status of a job can be modified by the user.

    Given an authenticated user and registered jobs, this test validates the following
    sequence of actions:

    - The user is able to modify the status of a registered job.
    - The returned job status matches the expected new status.
    """
    job_to_change_status = registered_jobs["job1"]
    job_id = job_to_change_status["id"]
    experiment_id = job_to_change_status["experiment"]["snapshotId"]
    new_status = "started"
    modify_job_status(
        client=client,
        job_id=job_id,
        experiment_id=experiment_id,
        new_status=new_status,
    )
    assert_job_status_matches_expectations(
        client, job_id=job_to_change_status["id"], expected=new_status
    )


def test_manage_job_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that different snapshots of a job can be retrieved by the user.

    Given an authenticated user and registered jobs, this test validates the following
    sequence of actions:

    - The user modifies a job status
    - The user retrieves information about the original snapshot of the job and gets
      the expected response
    - The user retrieves information about the new snapshot of the job and gets the
      expected response
    - The user retrieves a list of all snapshots of the job and gets the expected
      response
    """
    job_to_change_status = registered_jobs["job1"]
    job_id = job_to_change_status["id"]
    modify_job_status(
        client,
        job_id=job_to_change_status["id"],
        experiment_id=job_to_change_status["experiment"]["snapshotId"],
        new_status="started",
    )
    modified_job = get_job(client, job_id=job_id).get_json()
    modified_job.pop("hasDraft")
    modified_job.pop("artifacts")
    job_to_change_status.pop("hasDraft")
    job_to_change_status.pop("artifacts")
    job_to_change_status["latestSnapshot"] = False
    job_to_change_status["lastModifiedOn"] = modified_job["lastModifiedOn"]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=job_to_change_status["id"],
        snapshot_id=job_to_change_status["snapshot"],
        expected=job_to_change_status,
    )
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=modified_job["id"],
        snapshot_id=modified_job["snapshot"],
        expected=modified_job,
    )
    expected_snapshots = [job_to_change_status, modified_job]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=job_to_change_status["id"],
        expected=expected_snapshots,
    )


def test_tag_job(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to jobs.

    Given an authenticated user, registered jobs, and registered tags this test
    validates the following sequence of actions:

    - The user appends tags to a job
    - The user retrieves information about the tags of the job and gets
      the expected response
    - The user removes a tag from a job
    - The user retrieves information about the tags of the job and gets
      the expected response
    - The user modifies a tag from a job
    - The user retrieves information about the tags of the job and gets
      the expected response
    - The user removes all tags from a job
    - The user attempts to retrieve information about the tags of the job and gets
      the gets the expected response of no tags
    """
    job = registered_jobs["job1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=job["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=job["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client, resource_route=V1_JOBS_ROUTE, resource_id=job["id"], tag_id=tags[1]
    )
    response = actions.get_tags(
        client, resource_route=V1_JOBS_ROUTE, resource_id=job["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_JOBS_ROUTE,
        resource_id=job["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=V1_JOBS_ROUTE, resource_id=job["id"]
    )
    response = actions.get_tags(
        client, resource_route=V1_JOBS_ROUTE, resource_id=job["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(response.get_json(), [])



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import io
import tarfile
import time
from pathlib import Path
from typing import Any, BinaryIO, Iterable, List

import pytest
import requests
from boto3.session import Session
from botocore.client import BaseClient
from flask import Flask
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from injector import Binder, Injector
from mlflow.tracking import set_tracking_uri
from redis import Redis
from requests import ConnectionError
from requests import Session as RequestsSession

from dioptra.restapi.db import db as restapi_db
from dioptra.restapi.v1.shared.request_scope import request

from .lib import db as libdb
from .lib.server import FlaskTestServer


@pytest.fixture
def task_plugin_archive():
    archive_fileobj: BinaryIO = io.BytesIO()

    with (
        tarfile.open(fileobj=archive_fileobj, mode="w:gz") as f,
        io.BytesIO(initial_bytes=b"# init file") as f_init,
        io.BytesIO(initial_bytes=b"# plugin module") as f_plugin_module,
    ):
        tarinfo_init = tarfile.TarInfo(name="new_plugin_module/__init__.py")
        tarinfo_init.size = len(f_init.getbuffer())
        f.addfile(tarinfo=tarinfo_init, fileobj=f_init)

        tarinfo_plugin_module = tarfile.TarInfo(
            name="new_plugin_module/plugin_module.py"
        )
        tarinfo_plugin_module.size = len(f_plugin_module.getbuffer())
        f.addfile(tarinfo=tarinfo_plugin_module, fileobj=f_plugin_module)

    archive_fileobj.seek(0)
    yield archive_fileobj
    archive_fileobj.close()


@pytest.fixture
def dependency_modules() -> List[Any]:
    from dioptra.restapi.bootstrap import (
        PasswordServiceModule,
        RQServiceConfiguration,
        RQServiceV1Module,
        _bind_password_service_configuration,
    )

    def _bind_rq_service_configuration(binder: Binder) -> None:
        configuration: RQServiceConfiguration = RQServiceConfiguration(
            redis=Redis.from_url("redis://"),
        )

        binder.bind(
            interface=RQServiceConfiguration,
            to=configuration,
            scope=request,
        )

    def _bind_s3_service_configuration(binder: Binder) -> None:
        s3_session: Session = Session()
        s3_client: BaseClient = s3_session.client("s3")

        binder.bind(Session, to=s3_session, scope=request)
        binder.bind(BaseClient, to=s3_client, scope=request)

    def configure(binder: Binder) -> None:
        _bind_password_service_configuration(binder)
        _bind_rq_service_configuration(binder)
        _bind_s3_service_configuration(binder)

    return [
        configure,
        PasswordServiceModule,
        RQServiceV1Module,
    ]


@pytest.fixture
def dependency_injector(dependency_modules: List[Any]) -> Injector:
    return Injector(dependency_modules)


@pytest.fixture
def app(dependency_modules: List[Any]) -> Flask:
    from dioptra.restapi import create_app

    injector = Injector(dependency_modules)
    app = create_app(env="test_no_login", injector=injector)

    yield app


@pytest.fixture
def db(app: Flask) -> SQLAlchemy:
    with app.app_context():
        restapi_db.drop_all()
        restapi_db.create_all()
        yield restapi_db
        restapi_db.drop_all()
        restapi_db.session.commit()


@pytest.fixture(autouse=True)
def seed_database(db: SQLAlchemy):
    libdb.setup_ontology(db.session)


@pytest.fixture
def client(app: Flask) -> FlaskClient:
    return app.test_client()


@pytest.fixture
def flask_test_server(tmp_path: Path, http_client: RequestsSession):
    """Start a Flask test server.

    Args:
        tmp_path: The path to the temporary directory.
        http_client: A Requests session client.
    """
    sqlite_path = tmp_path / "dioptra-test.db"
    server = FlaskTestServer(sqlite_path=sqlite_path)
    server.upgrade_db()
    with server:
        wait_for_healthcheck_success(http_client)
        yield


@pytest.fixture
def http_client() -> Iterable[RequestsSession]:
    """A Requests session for accessing the API.

    Yields:
        A Requests session.
    """
    with requests.Session() as s:
        yield s


@pytest.fixture(autouse=True)
def create_mlruns(tmp_path):
    path = Path(tmp_path / "mlruns")
    path.mkdir()
    set_tracking_uri(path)


def wait_for_healthcheck_success(client: RequestsSession, timeout: int = 10) -> None:
    """Wait for the healthcheck endpoint to return a successful response.

    Args:
        client: A Requests session client.
        timeout: The maximum number of seconds to wait for the healthcheck endpoint to
            return a successful response.

    Raises:
        TimeoutError: If the healthcheck endpoint does not return a successful response
            within the specified timeout.
    """

    class StubResponse(object):
        """Stubbed response to use when the client experiences a connection error."""

        def json(self) -> str:
            """Return an empty string."""
            return ""

    healthcheck_url = "http://localhost:5000/health"
    time_elapsed = 0

    try:
        healthcheck = client.get(healthcheck_url)

    except ConnectionError:
        healthcheck = StubResponse()

    time_start = time.time()

    while healthcheck.json() != "healthy":
        if time_elapsed > timeout:
            raise TimeoutError("Healthcheck failed to return healthy status")

        time.sleep(0.1)

        try:
            healthcheck = client.get(healthcheck_url)

        except ConnectionError:
            healthcheck = StubResponse()

        time_elapsed = time.time() - time_start



"""empty message

Revision ID: 32d5c0055e1b
Revises: a278f46a1c9d
Create Date: 2020-11-18 17:45:26.337542

"""

import datetime

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import sessionmaker

try:
    from sqlalchemy.orm import declarative_base

except ImportError:
    from sqlalchemy.ext.declarative import declarative_base


# revision identifiers, used by Alembic.
revision = "32d5c0055e1b"
down_revision = "a278f46a1c9d"
branch_labels = None
depends_on = None

POSTGRES_DIALECT = "postgresql"

# Migration data models
Base = declarative_base()


class Job(Base):
    __tablename__ = "jobs"

    job_id = sa.Column(sa.String(36), primary_key=True)
    queue_id = sa.Column(sa.BigInteger(), sa.ForeignKey("queues.queue_id"), index=True)
    queue = sa.Column(sa.Text(), index=True)


class Queue(Base):
    __tablename__ = "queues"

    queue_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True
    )
    created_on = sa.Column(sa.DateTime())
    last_modified = sa.Column(sa.DateTime())
    name = sa.Column(sa.Text(), index=True, nullable=False, unique=True)
    is_deleted = sa.Column(sa.Boolean(), default=False)


def next_queue_id(session):
    """Generate the next id in the sequence."""
    stmt = sa.select(Queue).order_by(Queue.queue_id.desc())
    queue = session.scalars(stmt).first()

    if queue is None:
        return 1

    return int(queue.queue_id) + 1


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)
    dialect_name = op.get_context().dialect.name

    job_statuses = op.create_table(
        "job_statuses",
        sa.Column("status", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("status", name=op.f("pk_job_statuses")),
    )

    op.bulk_insert(
        job_statuses,
        [
            {"status": "queued"},
            {"status": "started"},
            {"status": "deferred"},
            {"status": "finished"},
            {"status": "failed"},
        ],
    )

    op.create_table(
        "queues",
        sa.Column(
            "queue_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("queue_id", name=op.f("pk_queues")),
    )

    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_queues_name"), ["name"], unique=True)

    op.create_table(
        "queue_locks",
        sa.Column(
            "queue_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["queue_id"],
            ["queues.queue_id"],
            name=op.f("fk_queue_locks_queue_id_queues"),
        ),
        sa.PrimaryKeyConstraint("queue_id", name=op.f("pk_queue_locks")),
    )

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.add_column(sa.Column("queue_id", sa.BigInteger(), nullable=True))
        batch_op.alter_column(
            "status",
            type_=sa.String(length=255),
            existing_type_=sa.Enum(
                "queued",
                "started",
                "deferred",
                "finished",
                "failed",
                name="jobstatus",
            ),
        )
        batch_op.create_index(
            batch_op.f("ix_jobs_queue_id"), ["queue_id"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_jobs_status"), ["status"], unique=False)
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_status_job_statuses"),
            "job_statuses",
            ["status"],
            ["status"],
        )
        batch_op.drop_index("ix_jobs_queue")
        batch_op.alter_column(
            "queue",
            type_=sa.String(length=255),
            existing_type_=sa.Enum(
                "tensorflow_cpu",
                "tensorflow_gpu",
                name="jobstatus",
            ),
        )

    with Session() as session:
        queue_name_id_dict = {}
        stmt = sa.select(Job)

        for job in session.scalars(stmt):
            if job.queue is not None and job.queue not in queue_name_id_dict:
                new_queue_id = next_queue_id(session)
                timestamp = datetime.datetime.now()

                new_queue = Queue(
                    queue_id=new_queue_id,
                    name=job.queue,
                    created_on=timestamp,
                    last_modified=timestamp,
                )
                session.add(new_queue)

                queue_name_id_dict[job.queue] = new_queue_id

            job.queue_id = queue_name_id_dict.get(job.queue)

        session.commit()

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_queue_id_queues"), "queues", ["queue_id"], ["queue_id"]
        )
        batch_op.drop_column("queue")

    if dialect_name == POSTGRES_DIALECT:
        op.execute("DROP TYPE IF EXISTS jobqueue; DROP TYPE IF EXISTS jobstatus;")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)
    dialect_name = op.get_context().dialect.name

    # Repeating this from the upgrade function since it was retroactively added. Will do
    # nothing if types have already been dropped.
    if dialect_name == POSTGRES_DIALECT:
        op.execute("DROP TYPE IF EXISTS jobqueue; DROP TYPE IF EXISTS jobstatus;")

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        if dialect_name == POSTGRES_DIALECT:
            op.execute(
                "CREATE TYPE jobqueue AS ENUM('tensorflow_cpu', 'tensorflow_gpu');"
            )

        batch_op.add_column(
            sa.Column(
                "queue",
                sa.Enum("tensorflow_cpu", "tensorflow_gpu", name="jobqueue"),
                nullable=True,
            ),
        )
        batch_op.drop_constraint(
            batch_op.f("fk_jobs_queue_id_queues"), type_="foreignkey"
        )
        batch_op.drop_constraint(
            batch_op.f("fk_jobs_status_job_statuses"), type_="foreignkey"
        )

    with Session() as session:
        downgraded_queue_names_set = {"tensorflow_cpu", "tensorflow_gpu"}
        stmt = sa.select(Queue.name, Job).join(Queue, Job.queue_id == Queue.queue_id)

        for queue_name, job in session.execute(stmt):
            job.queue = queue_name if queue_name in downgraded_queue_names_set else None

        session.commit()

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index("ix_jobs_queue", ["queue"], unique=False)
        batch_op.drop_index(batch_op.f("ix_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_jobs_queue_id"))

        if dialect_name == POSTGRES_DIALECT:
            op.execute(
                "CREATE TYPE jobstatus AS "
                "ENUM('queued', 'started', 'deferred', 'finished', 'failed'); "
                "ALTER TABLE jobs "
                "ALTER COLUMN status TYPE jobstatus "
                "USING status::text::jobstatus;"
            )

        else:
            batch_op.alter_column(
                "status",
                type_=sa.Enum(
                    "queued",
                    "started",
                    "deferred",
                    "finished",
                    "failed",
                    name="jobstatus",
                ),
                existing_type=sa.String(length=255),
            )

        batch_op.drop_column("queue_id")

    op.drop_table("queue_locks")
    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_queues_name"))

    op.drop_table("queues")
    op.drop_table("job_statuses")
    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, ForeignKeyConstraint, Index, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from dioptra.restapi.db.db import bigint, db, intpk, optionalstr, text_

from .resources import ResourceSnapshot

if TYPE_CHECKING:
    from .jobs import EntryPointJob
    from .plugins import Plugin, PluginFile
    from .resources import Resource


# -- Tables (no ORM) -------------------------------------------------------------------

entry_point_parameter_types_table = db.Table(
    "entry_point_parameter_types",
    Column("parameter_type", Text(), primary_key=True),
)

# -- ORM Classes -----------------------------------------------------------------------


class EntryPoint(ResourceSnapshot):
    __tablename__ = "entry_points"

    # Database fields
    resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(init=False, nullable=False, index=True)
    name: Mapped[text_] = mapped_column(nullable=False, index=True)
    task_graph: Mapped[text_] = mapped_column(nullable=False)

    # Relationships
    parameters: Mapped[list["EntryPointParameter"]] = relationship(
        back_populates="entry_point"
    )
    entry_point_jobs: Mapped[list["EntryPointJob"]] = relationship(
        init=False, viewonly=True
    )
    entry_point_plugin_files: Mapped[list["EntryPointPluginFile"]] = relationship(
        init=False, back_populates="entry_point"
    )

    # Additional settings
    __table_args__ = (  # type: ignore[assignment]
        Index(None, "resource_snapshot_id", "resource_id", unique=True),
        ForeignKeyConstraint(
            ["resource_snapshot_id", "resource_id"],
            [
                "resource_snapshots.resource_snapshot_id",
                "resource_snapshots.resource_id",
            ],
        ),
    )
    __mapper_args__ = {
        "polymorphic_identity": "entry_point",
    }


class EntryPointParameter(db.Model):  # type: ignore[name-defined]
    __tablename__ = "entry_point_parameters"

    # Database fields
    entry_point_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("entry_points.resource_snapshot_id"), init=False
    )
    parameter_number: Mapped[intpk]
    parameter_type: Mapped[text_] = mapped_column(
        ForeignKey("entry_point_parameter_types.parameter_type"),
        nullable=False,
        index=True,
    )
    name: Mapped[text_] = mapped_column(nullable=False)
    default_value: Mapped[optionalstr]

    # Relationships
    entry_point: Mapped["EntryPoint"] = relationship(
        init=False, back_populates="parameters"
    )
    values: Mapped[list["EntryPointParameterValue"]] = relationship(
        init=False, viewonly=True
    )

    __table_args__ = (
        Index(None, "entry_point_resource_snapshot_id", "name", unique=True),
    )


class EntryPointParameterValue(db.Model):  # type: ignore[name-defined]
    __tablename__ = "entry_point_parameter_values"

    # Database fields
    job_resource_id: Mapped[intpk] = mapped_column(
        ForeignKey("resources.resource_id"), init=False
    )
    entry_point_resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    parameter_number: Mapped[intpk] = mapped_column(init=False)
    value: Mapped[optionalstr]

    # Relationships
    job_resource: Mapped["Resource"] = relationship()
    parameter: Mapped["EntryPointParameter"] = relationship(
        back_populates="values", lazy="joined"
    )
    entry_point_job: Mapped["EntryPointJob"] = relationship(
        init=False,
        back_populates="entry_point_parameter_values",
        overlaps="job_resource,parameter,values",
    )

    # Additional settings
    __table_args__ = (
        Index(None, "job_resource_id", "parameter_number", unique=True),
        ForeignKeyConstraint(
            ["entry_point_resource_snapshot_id", "parameter_number"],
            [
                "entry_point_parameters.entry_point_resource_snapshot_id",
                "entry_point_parameters.parameter_number",
            ],
        ),
        ForeignKeyConstraint(
            ["entry_point_resource_snapshot_id", "job_resource_id"],
            [
                "entry_point_jobs.entry_point_resource_snapshot_id",
                "entry_point_jobs.job_resource_id",
            ],
        ),
    )


class EntryPointPluginFile(db.Model):  # type: ignore[name-defined]
    __tablename__ = "entry_point_plugin_files"

    # Database fields
    entry_point_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("entry_points.resource_snapshot_id"), init=False
    )
    plugin_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("plugins.resource_snapshot_id"), init=False
    )
    plugin_file_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("plugin_files.resource_snapshot_id"), init=False
    )

    # Relationships
    entry_point: Mapped["EntryPoint"] = relationship(
        back_populates="entry_point_plugin_files", lazy="joined"
    )
    plugin: Mapped["Plugin"] = relationship(lazy="joined")
    plugin_file: Mapped["PluginFile"] = relationship(lazy="joined")



"""Add ml_model_versions_resource

Revision ID: fd786b5377d6
Revises: d2bae5f6d991
Create Date: 2024-06-28 17:13:00.008695

"""

from typing import Annotated

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    sessionmaker,
)

# revision identifiers, used by Alembic.
revision = "fd786b5377d6"
down_revision = "d2bae5f6d991"
branch_labels = None
depends_on = None


# Migration data models
intpk = Annotated[
    int,
    mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True),
]
bigint = Annotated[
    int, mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"))
]
text_ = Annotated[str, mapped_column(sa.Text())]


class UpgradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class DowngradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class ResourceDependencyTypeUpgrade(UpgradeBase):
    __tablename__ = "resource_dependency_types"

    parent_resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), primary_key=True
    )
    child_resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), primary_key=True
    )


class ResourceTypeUpgrade(UpgradeBase):
    __tablename__ = "resource_types"

    resource_type: Mapped[text_] = mapped_column(primary_key=True)


class ResourceDependencyTypeDowngrade(DowngradeBase):
    __tablename__ = "resource_dependency_types"

    parent_resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), primary_key=True
    )
    child_resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), primary_key=True
    )


class ResourceTypeDowngrade(DowngradeBase):
    __tablename__ = "resource_types"

    resource_type: Mapped[text_] = mapped_column(primary_key=True)


class ResourceDependencyDowngrade(DowngradeBase):
    __tablename__ = "resource_dependencies"

    parent_resource_id: Mapped[intpk]
    child_resource_id: Mapped[intpk]
    parent_resource_type: Mapped[text_] = mapped_column(nullable=False)
    child_resource_type: Mapped[text_] = mapped_column(nullable=False)


class DraftResourceDowngrade(DowngradeBase):
    __tablename__ = "draft_resources"

    # Database fields
    draft_resource_id: Mapped[intpk]
    resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), nullable=False
    )


class SharedResourceDowngrade(DowngradeBase):
    __tablename__ = "shared_resources"

    # Database fields
    shared_resource_id: Mapped[intpk]
    resource_id: Mapped[bigint] = mapped_column(
        sa.ForeignKey("resources.resource_id"), nullable=False
    )


class ResourceDowngrade(DowngradeBase):
    __tablename__ = "resources"

    # Database fields
    resource_id: Mapped[intpk]
    resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), nullable=False
    )


class ResourceSnapshotDowngrade(DowngradeBase):
    __tablename__ = "resource_snapshots"

    # Database fields
    resource_snapshot_id: Mapped[intpk]
    resource_id: Mapped[bigint] = mapped_column(nullable=False)
    resource_type: Mapped[text_] = mapped_column(
        sa.ForeignKey("resource_types.resource_type"), nullable=False
    )


class MlModelDowngrade(DowngradeBase):
    __tablename__ = "ml_models"

    # Database fields
    resource_snapshot_id: Mapped[intpk]
    resource_id: Mapped[bigint] = mapped_column(nullable=False)


class MlModelVersionDowngrade(DowngradeBase):
    __tablename__ = "ml_model_versions"

    # Database fields
    resource_snapshot_id: Mapped[intpk]
    resource_id: Mapped[bigint] = mapped_column(nullable=False)


class ResourceTagDowngrade(DowngradeBase):
    __tablename__ = "resource_tags"

    # Database fields
    resource_id: Mapped[intpk] = mapped_column(sa.ForeignKey("resources.resource_id"))


def upgrade():
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    # Update the list of allowed resource types and resource dependency types
    with Session() as session:
        session.add(ResourceTypeUpgrade(resource_type="ml_model_version"))
        session.flush()
        session.add(
            ResourceDependencyTypeUpgrade(
                parent_resource_type="ml_model", child_resource_type="ml_model_version"
            )
        )
        session.commit()

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "ml_model_versions",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "artifact_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "version_number",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["artifact_resource_snapshot_id"],
            ["artifacts.resource_snapshot_id"],
            name=op.f("fk_ml_model_versions_artifact_resource_snapshot_id_artifacts"),
        ),
        sa.ForeignKeyConstraint(
            ["resource_snapshot_id", "resource_id"],
            [
                "resource_snapshots.resource_snapshot_id",
                "resource_snapshots.resource_id",
            ],
            name=op.f("fk_ml_model_versions_resource_snapshot_id_resource_snapshots"),
        ),
        sa.PrimaryKeyConstraint(
            "resource_snapshot_id", name=op.f("pk_ml_model_versions")
        ),
        sa.UniqueConstraint(
            "resource_snapshot_id",
            "resource_id",
            name=op.f("uq_ml_model_versions_resource_snapshot_id"),
        ),
    )
    with op.batch_alter_table("ml_model_versions", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_ml_model_versions_resource_id"),
            ["resource_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_ml_model_versions_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id", "artifact_resource_snapshot_id"],
            unique=True,
        )

    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_ml_models_resource_snapshot_id"))
        batch_op.create_index(
            batch_op.f("ix_ml_models_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )
        batch_op.drop_constraint(
            "fk_ml_models_artifact_resource_snapshot_id_artifacts", type_="foreignkey"
        )
        batch_op.drop_column("artifact_resource_snapshot_id")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    # Remove all traces of MlModel and MlModelVersion resources
    with Session() as session:
        ml_models_stmt = sa.select(MlModelDowngrade)
        ml_model_versions_stmt = sa.select(MlModelVersionDowngrade)
        drafts_stmt = sa.select(DraftResourceDowngrade).where(
            DraftResourceDowngrade.resource_type.in_(["ml_model", "ml_model_version"])
        )
        resources_stmt = sa.select(ResourceDowngrade).where(
            ResourceDowngrade.resource_type.in_(["ml_model", "ml_model_version"])
        )
        resource_snapshots_stmt = sa.select(ResourceSnapshotDowngrade).where(
            ResourceSnapshotDowngrade.resource_type.in_(
                ["ml_model", "ml_model_version"]
            )
        )
        resource_dependencies_stmt = sa.select(ResourceDependencyDowngrade).where(
            ResourceDependencyDowngrade.parent_resource_type == "ml_model",
            ResourceDependencyDowngrade.child_resource_type == "ml_model_version",
        )
        cte_resource_ids = (
            sa.select(ResourceDowngrade.resource_id)
            .where(
                ResourceDowngrade.resource_type.in_(["ml_model", "ml_model_version"])
            )
            .cte()
        )
        shared_resources_stmt = sa.select(SharedResourceDowngrade).where(
            SharedResourceDowngrade.resource_id.in_(sa.select(cte_resource_ids))
        )
        resource_tags_stmt = sa.select(ResourceTagDowngrade).where(
            ResourceTagDowngrade.resource_id.in_(sa.select(cte_resource_ids))
        )

        for ml_model in session.scalars(ml_models_stmt):
            session.delete(ml_model)

        for ml_model_version in session.scalars(ml_model_versions_stmt):
            session.delete(ml_model_version)

        for draft in session.scalars(drafts_stmt):
            session.delete(draft)

        for resource in session.scalars(resources_stmt):
            session.delete(resource)

        for resource_snapshot in session.scalars(resource_snapshots_stmt):
            session.delete(resource_snapshot)

        for resource_dependency in session.scalars(resource_dependencies_stmt):
            session.delete(resource_dependency)

        for shared_resource in session.scalars(shared_resources_stmt):
            session.delete(shared_resource)

        for resource_tag in session.scalars(resource_tags_stmt):
            session.delete(resource_tag)

        session.commit()

    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "artifact_resource_snapshot_id",
                sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
                nullable=True,
            )
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_ml_models_artifact_resource_snapshot_id_artifacts"),
            "artifacts",
            ["artifact_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.drop_index(batch_op.f("ix_ml_models_resource_snapshot_id"))
        batch_op.create_index(
            batch_op.f("ix_ml_models_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id", "artifact_resource_snapshot_id"],
            unique=True,
        )

    # Workaround to ensure the migration won't fail (table should be empty)
    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.alter_column("artifact_resource_snapshot_id", nullable=False)

    with op.batch_alter_table("ml_model_versions", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_ml_model_versions_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_ml_model_versions_resource_id"))

    op.drop_table("ml_model_versions")

    # Downgrade the list of allowed resource types and resource dependency types
    with Session() as session:
        ml_model_version_type_stmt = sa.select(ResourceTypeDowngrade).where(
            ResourceTypeDowngrade.resource_type == "ml_model_version"
        )
        resource_dependency_types_stmt = sa.select(
            ResourceDependencyTypeDowngrade
        ).where(
            ResourceDependencyTypeDowngrade.parent_resource_type == "ml_model",
            ResourceDependencyTypeDowngrade.child_resource_type == "ml_model_version",
        )
        ml_model_version_type = session.scalar(ml_model_version_type_stmt)
        resource_dependency_type = session.scalar(resource_dependency_types_stmt)

        if ml_model_version_type is not None:
            session.delete(ml_model_version_type)

        if resource_dependency_type is not None:
            session.delete(resource_dependency_type)

        session.commit()

    # ### end Alembic commands ###



"""Add 'readonly' resource locks to preexisting built-in parameter types

Revision ID: d2bae5f6d991
Revises: 10f9e72e72aa
Create Date: 2024-06-28 12:30:13.685590

"""

import datetime
from typing import Annotated, Any, Optional

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    column_property,
    mapped_column,
    relationship,
    sessionmaker,
)

from dioptra.restapi.db.custom_types import TZDateTime

# revision identifiers, used by Alembic.
revision = "d2bae5f6d991"
down_revision = "10f9e72e72aa"
branch_labels = None
depends_on = None


# Declare list of allowed simple built-in parameter types
BUILTIN_PARAMETER_TYPES = ["any", "number", "integer", "string", "boolean", "null"]


# Migration data models
bigint = Annotated[
    int, mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"))
]
intpk = Annotated[
    int,
    mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True),
]
text_ = Annotated[str, mapped_column(sa.Text())]
bool_ = Annotated[bool, mapped_column(sa.Boolean())]
datetimetz = Annotated[datetime.datetime, mapped_column(TZDateTime())]
optionalbigint = Annotated[
    Optional[int],
    mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), nullable=True),
]
optionaljson_ = Annotated[Optional[dict[str, Any]], mapped_column(sa.JSON)]
optionalstr = Annotated[Optional[str], mapped_column(sa.Text(), nullable=True)]


class UpgradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class GroupUpgrade(UpgradeBase):
    __tablename__ = "groups"

    # Database fields
    group_id: Mapped[intpk] = mapped_column(init=False)
    name: Mapped[text_] = mapped_column(nullable=False)
    user_id: Mapped[bigint] = mapped_column(nullable=False)


class ResourceSnapshotUpgrade(UpgradeBase):
    __tablename__ = "resource_snapshots"

    # Database fields
    resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(nullable=False)
    resource_type: Mapped[text_] = mapped_column(nullable=False)
    user_id: Mapped[bigint] = mapped_column(nullable=False)
    description: Mapped[optionalstr] = mapped_column()
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class ResourceLockUpgrade(UpgradeBase):
    __tablename__ = "resource_locks"

    # Database fields
    resource_id: Mapped[intpk] = mapped_column(sa.ForeignKey("resources.resource_id"))
    resource_lock_type: Mapped[text_] = mapped_column(primary_key=True)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    resource: Mapped["ResourceUpgrade"] = relationship(
        init=False, back_populates="locks"
    )

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class ResourceUpgrade(UpgradeBase):
    __tablename__ = "resources"

    # Database fields
    resource_id: Mapped[intpk] = mapped_column(init=False)
    group_id: Mapped[bigint] = mapped_column(nullable=False)
    resource_type: Mapped[text_] = mapped_column(nullable=False)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Derived fields (read-only)
    is_deleted: Mapped[bool] = column_property(
        sa.select(ResourceLockUpgrade.resource_id)
        .where(
            ResourceLockUpgrade.resource_id == resource_id,
            ResourceLockUpgrade.resource_lock_type == "delete",
        )
        .correlate_except(ResourceLockUpgrade)
        .exists()
    )
    is_readonly: Mapped[bool] = column_property(
        sa.select(ResourceLockUpgrade.resource_id)
        .where(
            ResourceLockUpgrade.resource_id == resource_id,
            ResourceLockUpgrade.resource_lock_type == "readonly",
        )
        .correlate_except(ResourceLockUpgrade)
        .exists()
    )
    latest_snapshot_id: Mapped[optionalbigint] = column_property(
        sa.Nullable(
            sa.select(ResourceSnapshotUpgrade.resource_snapshot_id)
            .where(ResourceSnapshotUpgrade.resource_id == resource_id)
            .order_by(ResourceSnapshotUpgrade.created_on.desc())
            .limit(1)
            .correlate_except(ResourceSnapshotUpgrade)
            .scalar_subquery()
        )
    )

    # Relationships
    locks: Mapped[list["ResourceLockUpgrade"]] = relationship(
        init=False, back_populates="resource"
    )

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class PluginTaskParameterTypeUpgrade(UpgradeBase):
    __tablename__ = "plugin_task_parameter_types"

    # Database fields
    resource_snapshot_id: Mapped[intpk]
    resource_id: Mapped[bigint] = mapped_column(nullable=False)
    name: Mapped[text_] = mapped_column(nullable=False)
    structure: Mapped[optionaljson_] = mapped_column(nullable=True)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with Session() as session:
        # Search for built-in simple types that are in the database already and haven't
        # been marked as deleted and don't have a "readonly" lock set. Set a "readonly"
        # lock on any records found that match this criteria.
        param_types_no_readonly_locks_stmt = (
            sa.select(PluginTaskParameterTypeUpgrade.resource_id)
            .join(
                ResourceUpgrade,
                PluginTaskParameterTypeUpgrade.resource_id
                == ResourceUpgrade.resource_id,
            )
            .where(
                PluginTaskParameterTypeUpgrade.name.in_(BUILTIN_PARAMETER_TYPES),
                ResourceUpgrade.is_deleted == False,  # noqa: E712
                ResourceUpgrade.is_readonly == False,  # noqa: E712
                ResourceUpgrade.latest_snapshot_id
                == PluginTaskParameterTypeUpgrade.resource_snapshot_id,
            )
        )

        for resource_id in session.scalars(param_types_no_readonly_locks_stmt):
            resource_lock = ResourceLockUpgrade(
                resource_id=resource_id, resource_lock_type="readonly"
            )
            session.add(resource_lock)

        # Get information about the default public group.
        public_group = session.scalar(
            sa.select(GroupUpgrade).where(GroupUpgrade.name == "public")
        )

        # If the public group already exists, then get a list of the built-in simple
        # types that are already registered in the public group. Figure out which
        # built-in types are still missing, register them, and mark them as readonly.
        if public_group:
            existing_builtin_param_types_stmt = (
                sa.select(PluginTaskParameterTypeUpgrade.name)
                .join(
                    ResourceUpgrade,
                    PluginTaskParameterTypeUpgrade.resource_id
                    == ResourceUpgrade.resource_id,
                )
                .where(
                    PluginTaskParameterTypeUpgrade.name.in_(BUILTIN_PARAMETER_TYPES),
                    ResourceUpgrade.group_id == public_group.group_id,
                    ResourceUpgrade.is_deleted == False,  # noqa: E712
                    ResourceUpgrade.latest_snapshot_id
                    == PluginTaskParameterTypeUpgrade.resource_snapshot_id,
                )
            )
            existing_builtin_param_types = list(
                session.scalars(existing_builtin_param_types_stmt).all()
            )
            missing_builtin_param_types = [
                x
                for x in BUILTIN_PARAMETER_TYPES
                if x not in set(existing_builtin_param_types)
            ]

            for builtin_type in missing_builtin_param_types:
                resource = ResourceUpgrade(
                    group_id=public_group.group_id,
                    resource_type="plugin_task_parameter_type",
                )
                session.add(resource)
                session.flush()
                resource_snapshot = ResourceSnapshotUpgrade(
                    resource_id=resource.resource_id,
                    resource_type="plugin_task_parameter_type",
                    user_id=public_group.user_id,
                    description=None,
                )
                session.add(resource_snapshot)
                session.flush()
                param_type = PluginTaskParameterTypeUpgrade(
                    resource_snapshot_id=resource_snapshot.resource_snapshot_id,
                    resource_id=resource.resource_id,
                    name=builtin_type,
                    structure=None,
                )
                resource_lock = ResourceLockUpgrade(
                    resource_id=resource.resource_id, resource_lock_type="readonly"
                )
                session.add(param_type)
                session.add(resource_lock)
                session.flush()

        session.commit()


def downgrade():
    # No downgrade necessary, the readonly locks are simply ignored and hidden when
    # using previous commits.
    pass



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Tuple

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.bounding_boxes.iou import (
    TensorflowBoundingBoxesBatchedGridIOU,
)

from .classification import YOLOV1ClassificationLoss
from .localization import YOLOV1LocalizationLoss

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor
    from tensorflow.keras.losses import Loss

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class YOLOV1Loss(Loss):
    def __init__(
        self,
        bbox_grid_iou: TensorflowBoundingBoxesBatchedGridIOU,
        coord_weight: float = 5.0,
        object_weight: float = 5.0,
        no_object_weight: float = 1.0,
        labels_weight: float = 1.0,
        wh_loss_type: str = "sq_relative_diff",
        classification_loss_type: str = "categorical_crossentropy",
        name: str = "yolov1_loss",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self._bbox_grid_iou = bbox_grid_iou
        self._coord_weight = coord_weight
        self._object_weight = object_weight
        self._no_object_weight = no_object_weight
        self._labels_weight = labels_weight
        self._wh_loss_type = wh_loss_type
        self._classification_loss_type = classification_loss_type

        self._classification_loss = YOLOV1ClassificationLoss(
            labels_weight=labels_weight, loss_type=classification_loss_type
        )
        self._localization_loss = YOLOV1LocalizationLoss(
            coord_weight=coord_weight,
            object_weight=object_weight,
            no_object_weight=no_object_weight,
            wh_loss_type=wh_loss_type,
        )

    def call(self, y_true, y_pred):
        true_bboxes_cell_xywh = y_true[0]
        true_labels = y_true[1]
        true_object = y_true[2]
        true_no_object = y_true[3]

        pred_bboxes_cell_xywh = y_pred[0]
        pred_object_conf = y_pred[1]
        pred_labels = y_pred[2]

        (
            responsible_pred_bboxes_cell_xywh,
            responsible_pred_object_conf,
            not_responsible_pred_object_conf,
        ) = self._select_responsible_bboxes(
            true_bboxes_cell_xywh=true_bboxes_cell_xywh,
            pred_bboxes_cell_xywh=pred_bboxes_cell_xywh,
            pred_object_conf=pred_object_conf,
        )
        true_object_conf = self._compute_true_object_conf(
            true_bboxes_cell_xywh=true_bboxes_cell_xywh,
            pred_bboxes_cell_xywh=pred_bboxes_cell_xywh,
            true_object=true_object,
        )

        # Compute localization loss
        localization_loss = self._localization_loss(
            responsible_pred_bboxes_cell_xywh=responsible_pred_bboxes_cell_xywh,
            pred_object_conf=pred_object_conf,
            responsible_pred_object_conf=responsible_pred_object_conf,
            not_responsible_pred_object_conf=not_responsible_pred_object_conf,
            true_bboxes_cell_xywh=true_bboxes_cell_xywh,
            true_object_conf=true_object_conf,
            true_object=true_object,
            true_no_object=true_no_object,
        )

        # Compute classification loss
        classification_loss = self._classification_loss(
            pred_labels=pred_labels,
            true_labels=true_labels,
            true_object=true_object,
        )

        # Compute total loss
        yolo_loss = localization_loss + classification_loss

        return yolo_loss

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_true_object_conf(
        self,
        true_bboxes_cell_xywh: Tensor,
        pred_bboxes_cell_xywh: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        pred_bbox_max_iou = self._bbox_grid_iou.max_iou(
            bboxes_cell_xywh1=pred_bboxes_cell_xywh,
            bboxes_cell_xywh2=true_bboxes_cell_xywh,
        )

        return pred_bbox_max_iou * true_object

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _select_responsible_bboxes(
        self,
        true_bboxes_cell_xywh: Tensor,
        pred_bboxes_cell_xywh: Tensor,
        pred_object_conf: Tensor,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        (
            responsible_pred_bboxes_cell_xywh,
            responsible_pred_object_conf,
            not_responsible_pred_object_conf,
        ) = self._bbox_grid_iou.select_max_iou_bboxes(
            bboxes_cell_xywh=pred_bboxes_cell_xywh,
            bboxes_conf=pred_object_conf,
            bboxes_cell_xywh_ground_truth=true_bboxes_cell_xywh,
        )
        responsible_pred_bboxes_cell_xywh = tf.expand_dims(
            responsible_pred_bboxes_cell_xywh, axis=-2
        )

        return (
            responsible_pred_bboxes_cell_xywh,
            responsible_pred_object_conf,
            not_responsible_pred_object_conf,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import argparse
import logging
import logging.config
from collections.abc import Iterable
from typing import Any, Union

import yaml

import dioptra.task_engine.task_engine


def _parse_args() -> argparse.Namespace:
    """
    Set up and parse commandline parameters.
    """
    arg_parser = argparse.ArgumentParser(
        description="Read/process declarative experiment description",
    )

    arg_parser.add_argument(
        "file",
        type=argparse.FileType(mode="r", encoding="utf-8"),
        help="The declarative experiment file to read",
    )

    arg_parser.add_argument(
        "-P",
        help="""
        Assign values to task graph global parameters.  Values must be of the
        form <name>=<value>.  This option can be repeated.
        """,
        action="append",
        metavar="name=value",
    )

    arg_parser.add_argument(
        "-l",
        "--log-level",
        help="""
        Set the logging level.  "all" is a special setting which causes logging
        from all modules to display at the debug level, in addition to this
        application.  This can result in a lot of logging noise.
        Default: %(default)s
        """,
        choices=["all", "debug", "info", "warning", "error", "critical"],
        default="info",
    )

    return arg_parser.parse_args()


def _setup_logging(log_level: Union[int, str] = logging.INFO) -> None:
    """
    Set up logging.

    Args:
        log_level: The logging level to use.  May be one of the integer log
            level constants or names recognized by the logging module, or
            "all".
    """

    if isinstance(log_level, str):
        log_level = log_level.upper()

    config = {
        "version": 1,
        "formatters": {
            "fmt": {
                # Uniform width, so we get nice alignment, with enough room for
                # the word "critical".
                "format": "[%(levelname)8s] %(message)s"
            }
        },
        "handlers": {"stream": {"class": "logging.StreamHandler", "formatter": "fmt"}},
        "disable_existing_loggers": False,
    }

    logger_config: dict[str, Any] = {"handlers": ["stream"]}

    # Set the logging config on the root logger if "ALL", else just on
    # the task_graph module.
    if log_level == "ALL":
        logger_config["level"] = "DEBUG"
        config["root"] = logger_config
    else:
        logger_config["level"] = log_level
        config["loggers"] = {"dioptra.task_engine": logger_config}

    logging.config.dictConfig(config)


def _cmdline_params_to_map(params: Iterable[str]) -> dict[str, str]:
    """
    Convert global parameters given on the commandline in <name>=<value>
    format, into a mapping from name to value.  Values are run through a YAML
    evaluation so that there is more flexibility in terms of value types, than
    just using strings.  If an equals sign and value are missing (e.g. it is
    just <name>), this is treated the same as <name>=true.  If the format is
    =<value> this is an error: a parameter name is required.  If the format is
    <name>= this is not an error: the value is the empty string.

    Args:
        params: The commandline parameters defining global experiment
            parameters, as collected by argparse.

    Returns:
        A mapping from name (string) to value (some python type)
    """
    param_value: Any
    param_map = {}

    if params:
        for param in params:
            equals_idx = param.find("=")

            if equals_idx == 0:
                raise ValueError("Global parameter names must not be empty: " + param)

            elif equals_idx < 0:
                # No equals sign: treat this as setting a "true" value.
                param_name = param
                param_value = True

            else:
                param_name = param[:equals_idx]
                param_value = param[equals_idx + 1 :]

                # run a yaml evaluation to try to convert the value to some
                # Python type which is more specialized than a string.
                param_value = yaml.safe_load(param_value)

            param_map[param_name] = param_value

    return param_map


def main() -> None:
    args = _parse_args()
    _setup_logging(args.log_level)

    experiment_desc = yaml.safe_load(args.file)
    global_parameters = _cmdline_params_to_map(args.P)

    dioptra.task_engine.task_engine.run_experiment(experiment_desc, global_parameters)


if __name__ == "__main__":
    main()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# This is a fork of the work
# https://github.com/gahjelle/pyplugs/blob/90e635777672f75080291c737f08453a26ea380d/pyplugs/_plugins.py  # noqa: B950
# See copyright below.
#
# Copyright (c) 2019 Geir Arne Hjelle
# Distributed under the terms of the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Decorators for registering plugins"""

from __future__ import annotations

import functools
import importlib
import sys
import textwrap
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.exceptions import (
    PrefectDependencyError,
    UnknownPackageError,
    UnknownPluginError,
    UnknownPluginFunctionError,
)
from dioptra.sdk.utilities.decorators import require_package

LOGGER: BoundLogger = structlog.stdlib.get_logger()


try:
    from importlib import resources

except ImportError:  # pragma: nocover
    import importlib_resources as resources  # type: ignore

try:
    from prefect import task

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="prefect",
    )

try:
    from typing import Protocol

except ImportError:  # pragma: nocover
    from typing_extensions import Protocol  # type: ignore

if TYPE_CHECKING:
    from prefect.tasks.core.function import FunctionTask


# Structural subtyping
class NoutPlugin(Protocol):
    _task_nout: int

    def __call__(self, *args, **kwargs) -> Any: ...  # noqa: E704; pragma: nocover


# Type aliases
T = TypeVar("T")
Plugin = Callable[..., Any]


class PluginInfo(NamedTuple):
    """Information about one plug-in"""

    package_name: str
    plugin_name: str
    func_name: str
    func: Union[Plugin, NoutPlugin]
    description: str
    doc: str
    module_doc: str
    sort_value: float


# Dictionary with information about all registered plug-ins
_PLUGINS: Dict[str, Dict[str, Dict[str, PluginInfo]]] = {}


@overload
def register(*, sort_value: float) -> Callable[[Plugin], Plugin]:
    """Signature for using decorator with parameters"""
    ...  # pragma: nocover


@overload
def register(func: Plugin) -> Plugin:
    """Signature for using decorator without parameters"""
    ...  # pragma: nocover


def register(_func=None, *, sort_value=0):
    """Decorator for registering a new plug-in"""

    def decorator_register(func: Callable[..., T]) -> Callable[..., T]:
        """Store information about the given function"""
        package_name, _, plugin_name = func.__module__.rpartition(".")
        description, _, doc = (func.__doc__ or "").partition("\n\n")
        func_name = func.__name__
        module_doc = sys.modules[func.__module__].__doc__ or ""

        pkg_info = _PLUGINS.setdefault(package_name, {})
        plugin_info = pkg_info.setdefault(plugin_name, {})
        plugin_info[func_name] = PluginInfo(
            package_name=package_name,
            plugin_name=plugin_name,
            func_name=func_name,
            func=func,
            description=description,
            doc=textwrap.dedent(doc).strip(),
            module_doc=module_doc,
            sort_value=sort_value,
        )

        return func

    if _func is None:
        return decorator_register

    else:
        return decorator_register(_func)


def task_nout(nout: int) -> Callable[[Plugin], NoutPlugin]:
    def decorator(func: Plugin) -> NoutPlugin:
        # We're just assigning an attribute, and we need mypy to let us
        # do that.  So we just force a type change, to a callable type which
        # includes an attribute.
        nout_func = cast(NoutPlugin, func)
        nout_func._task_nout = nout

        return nout_func

    return decorator


def names(package: str) -> List[str]:
    """List all plug-ins in one package"""
    _import_all(package)

    return sorted(_PLUGINS[package].keys(), key=lambda p: info(package, p).sort_value)


def funcs(package: str, plugin: str) -> List[str]:
    """List all functions in one plug-in"""
    _import(package, plugin)
    plugin_info = _PLUGINS[package][plugin]

    return list(plugin_info.keys())


def info(package: str, plugin: str, func: Optional[str] = None) -> PluginInfo:
    """Get information about a plug-in"""
    _import(package, plugin)

    try:
        plugin_info = _PLUGINS[package][plugin]

    except KeyError as exc:
        raise UnknownPluginError(
            f"Could not find any plug-in named {plugin!r} inside {package!r}. "
            "Use pyplugs.register to register functions as plug-ins"
        ) from exc

    func = next(iter(plugin_info.keys())) if func is None else func

    try:
        return plugin_info[func]

    except KeyError as exc:
        raise UnknownPluginFunctionError(
            f"Could not find any function named {func!r} inside '{package}.{plugin}'. "
            "Use pyplugs.register to register plug-in functions"
        ) from exc


def exists(package: str, plugin: str) -> bool:
    """Check if a given plugin exists"""
    if package in _PLUGINS and plugin in _PLUGINS[package]:
        return True

    try:
        _import(package, plugin)

    except (UnknownPluginError, UnknownPackageError):
        return False

    else:
        return package in _PLUGINS and plugin in _PLUGINS[package]


def get(package: str, plugin: str, func: Optional[str] = None) -> Plugin:
    """Get a given plugin"""
    return info(package, plugin, func).func


def call(
    package: str, plugin: str, func: Optional[str] = None, *args: Any, **kwargs: Any
) -> Any:
    """Call the given plugin"""
    plugin_func = get(package, plugin, func)

    return plugin_func(*args, **kwargs)


@require_package("prefect", exc_type=PrefectDependencyError)
def get_task(package: str, plugin: str, func: Optional[str] = None) -> FunctionTask:
    """Get a given plugin wrapped as a prefect task"""
    plugin_func: Union[Plugin, NoutPlugin] = info(package, plugin, func).func
    nout: Optional[int] = getattr(plugin_func, "_task_nout", None)

    return task(nout=nout)(plugin_func)


@require_package("prefect", exc_type=PrefectDependencyError)
def call_task(
    package: str, plugin: str, func: Optional[str] = None, *args: Any, **kwargs: Any
) -> Any:
    """Call the given plugin as a prefect task"""
    plugin_task = get_task(package, plugin, func)

    return plugin_task(*args, **kwargs)


def _import(package: str, plugin: str) -> None:
    """Import the given plugin file from a package"""
    if package in _PLUGINS and plugin in _PLUGINS[package]:
        return None

    # If the plugin does not have a package, it is registered with package ""
    # (empty string).  Don't prepend the package in that case.
    if package:
        plugin_module = f"{package}.{plugin}"
    else:
        plugin_module = plugin

    try:
        importlib.import_module(plugin_module)

    except ImportError as err:
        if repr(plugin_module) in err.msg:
            #   Plugin not found in ''
            # is a confusing error message... don't mention a package in the
            # error message if there wasn't one.
            message = f"Plugin {plugin!r} not found"
            if package:
                message += f" in {package!r}"
            raise UnknownPluginError(message) from None

        elif repr(package) in err.msg:
            raise UnknownPackageError(f"Package {package!r} does not exist") from None

        raise


def _import_all(package: str) -> None:
    """Import all plugins in a package"""
    try:
        all_resources = resources.contents(package)

    except ImportError as err:
        raise UnknownPackageError(err) from None

    # Note that we have tried to import the package by adding it to _PLUGINS
    _PLUGINS.setdefault(package, {})

    # Loop through all Python files in the directories of the package
    plugins = [
        r[:-3] for r in all_resources if r.endswith(".py") and not r.startswith("_")
    ]

    for plugin in plugins:
        try:
            _import(package, plugin)

        except ImportError:
            pass  # Don't let errors in one plugin, affect the others


def names_factory(package: str) -> Callable[[], List[str]]:
    """Create a names() function for one package"""
    return functools.partial(names, package)


def funcs_factory(package: str) -> Callable[[str], List[str]]:
    """Create a funcs() function for one package"""
    return functools.partial(funcs, package)


def info_factory(package: str) -> Callable[[str, Optional[str]], PluginInfo]:
    """Create a info() function for one package"""
    return functools.partial(info, package)


def exists_factory(package: str) -> Callable[[str], bool]:
    """Create an exists() function for one package"""
    return functools.partial(exists, package)


def get_factory(package: str) -> Callable[[str, Optional[str]], Plugin]:
    """Create a get() function for one package"""
    return functools.partial(get, package)


def call_factory(package: str) -> Callable[..., Any]:
    """Create a call() function for one package"""
    return functools.partial(call, package)


@require_package("prefect", exc_type=PrefectDependencyError)
def get_task_factory(package: str) -> Callable[[str, Optional[str]], FunctionTask]:
    """Create a get_task() function for one package"""
    return functools.partial(get_task, package)


@require_package("prefect", exc_type=PrefectDependencyError)
def call_task_factory(package: str) -> Callable[..., Any]:
    """Create a call_task() function for one package"""
    return functools.partial(call_task, package)


__all__ = [
    "register",
    "task_nout",
    "names",
    "funcs",
    "info",
    "exists",
    "get",
    "call",
    "get_task",
    "call_task",
    "names_factory",
    "funcs_factory",
    "info_factory",
    "exists_factory",
    "get_factory",
    "call_factory",
    "get_task_factory",
    "call_task_factory",
]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Command-line script for applying migrations to the Dioptra database.

The database connection details are set using environment variables, see
dioptra/restapi/config.py for more information.

Examples:
    To upgrade the database to the latest revision::

        dioptra-db autoupgrade

    To downgrade the database to a previous revision::

        # Assume previous revision has a hash of 6b432d0a5ac7
        dioptra-db downgrade --revision 6b432d0a5ac7

    To display the database schema history::

        dioptra-db history

    To generate a new migration script::

        # Execute from the root of the repository
        dioptra-db migrate --message "Add new table"
"""
from __future__ import annotations

import os
import subprocess
from importlib.util import find_spec
from pathlib import Path

import click

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def _flask_db_current(alembic_dir: Path) -> str | None:
    """Get the current revision of the database.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.

    Returns:
        The current revision of the database.
    """
    p = subprocess.run(
        ["flask", "db", "current", "-d", str(alembic_dir)],
        capture_output=True,
        text=True,
    )
    output = p.stdout.strip().split(" ")[0]

    return output or None


def _flask_db_heads(alembic_dir: Path) -> str | None:
    """Get the HEAD revision of the database.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.

    Returns:
        The head revision of the database.
    """
    p = subprocess.run(
        ["flask", "db", "heads", "-d", str(alembic_dir)],
        capture_output=True,
        text=True,
    )
    output = p.stdout.strip().split(" ")[0]

    return output or None


def _flask_db_downgrade(alembic_dir: Path, revision: str) -> None:
    """Downgrade the database to the specified revision.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
        revision: The revision identifier.
    """
    args = ["flask", "db", "downgrade", "-d", str(alembic_dir), revision]

    subprocess.run(
        args=args,
        capture_output=False,
        text=True,
    )


def _flask_db_history(alembic_dir: Path, indicate_current: bool) -> None:
    """Display the database schema history.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
        indicate_current: Indicate the current revision in the history.
    """
    args = ["flask", "db", "history", "-d", str(alembic_dir)]

    if indicate_current:
        args.append("--indicate-current")

    subprocess.run(
        args=args,
        capture_output=False,
        text=True,
    )


def _flask_db_upgrade(alembic_dir: Path, revision: str) -> None:
    """Upgrade the database to the specified revision.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
        revision: The revision identifier.
    """
    args = ["flask", "db", "upgrade", "-d", str(alembic_dir), revision]
    subprocess.run(
        args=args,
        capture_output=False,
        text=True,
    )


def _flask_db_migrate(
    alembic_dir: Path,
    message: str | None,
    sql: bool,
    head: str | None,
    splice: bool,
    branch_label: str | None,
    version_path: str | None,
    rev_id: str | None,
    x_arg: list[str],
) -> None:
    """Autogenerate a new revision file.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
        message: The revision message.
        sql: Don't emit SQL to database - dump to standard output instead.
        head: Specify head revision or <branchname>@head to base new revision on.
        splice: Allow a non-head revision as the "head" to splice onto.
        branch_label: Specify a branch label to apply to the new revision.
        version_path: Specify specific path from config for version file.
        rev_id: Specify a hardcoded revision id instead of generating one.
        x_arg: Additional arguments consumed by custom env.py scripts.
    """
    args = ["flask", "db", "migrate", "-d", str(alembic_dir)]

    if message is not None:
        args.extend(["--message", message])

    if sql:
        args.append("--sql")

    if head is not None:
        args.extend(["--head", head])

    if splice:
        args.append("--splice")

    if branch_label is not None:
        args.extend(["--branch-label", branch_label])

    if version_path is not None:
        args.extend(["--version-path", version_path])

    if rev_id is not None:
        args.extend(["--rev-id", rev_id])

    if x_arg:
        for arg in x_arg:
            args.extend(["--x-arg", arg])

    subprocess.run(
        args=args,
        capture_output=False,
        text=True,
    )


def _flask_db_revision(
    alembic_dir: Path,
    message: str | None,
    autogenerate: bool,
    sql: bool,
    head: str | None,
    splice: bool,
    branch_label: str | None,
    version_path: str | None,
    rev_id: str | None,
) -> None:
    """Autogenerate a new revision file.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
        message: The revision message.
        autogenerate: Populate revision script with candidate migration operations,
            based on comparison of database to model.
        sql: Don't emit SQL to database - dump to standard output instead.
        head: Specify head revision or <branchname>@head to base new revision on.
        splice: Allow a non-head revision as the "head" to splice onto.
        branch_label: Specify a branch label to apply to the new revision.
        version_path: Specify specific path from config for version file.
        rev_id: Specify a hardcoded revision id instead of generating one.
    """
    args = ["flask", "db", "revision", "-d", str(alembic_dir)]

    if message is not None:
        args.extend(["--message", message])

    if autogenerate:
        args.append("--autogenerate")

    if sql:
        args.append("--sql")

    if head is not None:
        args.extend(["--head", head])

    if splice:
        args.append("--splice")

    if branch_label is not None:
        args.extend(["--branch-label", branch_label])

    if version_path is not None:
        args.extend(["--version-path", version_path])

    if rev_id is not None:
        args.extend(["--rev-id", rev_id])

    subprocess.run(
        args=args,
        capture_output=False,
        text=True,
    )


def autoupgrade_cmd(alembic_dir: Path) -> None:
    """Command for automatically upgrading the database schema to the latest version.

    Args:
        alembic_dir: Path to directory containing the Alembic migration scripts.
    """
    db_revision_id = _flask_db_current(alembic_dir=alembic_dir)
    head_revision_id = _flask_db_heads(alembic_dir=alembic_dir)

    if db_revision_id == head_revision_id:
        print("Database is up to date.")
        return None

    if db_revision_id is None:
        print("Initializing database.")
        _flask_db_upgrade(alembic_dir=alembic_dir, revision="head")
        return None

    print("Migrating database tables to latest revision.")
    _flask_db_upgrade(alembic_dir=alembic_dir, revision="head")


def find_alembic_dir() -> Path:
    """Find the directory containing the Alembic migration scripts.

    Returns:
        The path to the directory containing the Alembic migration scripts.
    """
    module_spec = find_spec("dioptra.restapi.db.alembic")

    if module_spec is None or module_spec.submodule_search_locations is None:
        raise ModuleNotFoundError(
            "Unable to find the spec for the dioptra.restapi.db.alembic submodule. "
            "The dioptra package does not appear to be installed correctly."
        )

    alembic_dir: Path = Path(module_spec.submodule_search_locations[0])
    return alembic_dir


def is_db_uri_set() -> bool:
    """Check if the database URI is set in the environment.

    Returns:
        True if the database URI is set in the environment, False otherwise.
    """
    app_env = os.getenv("DIOPTRA_RESTAPI_ENV", "dev")
    env_vars = {
        "dev": "DIOPTRA_RESTAPI_DEV_DATABASE_URI",
        "prod": "DIOPTRA_RESTAPI_DATABASE_URI",
    }

    if (env_var := env_vars.get(app_env)) is None:
        raise ValueError(
            f"Invalid value for DIOPTRA_RESTAPI_ENV: {app_env}. "
            f"Valid values are 'dev' and 'prod'."
        )

    result = os.getenv(env_var) is not None

    if not result:
        print(
            f"Environment variable {env_var} is not set. "
            "The database cannot be found. Please set the environment variable and "
            "try again."
        )

    return result


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """Command-line interface for managing the Dioptra database.

    The database connection details are set using the following environment
    variables.

    \b
    DIOPTRA_RESTAPI_ENV:
        The environment in which the Dioptra REST API is running. Valid values
        are 'dev' and 'prod'. Defaults to 'dev'.

    \b
    DIOPTRA_RESTAPI_DEV_DATABASE_URI:
        The connection URI for the development database. Used when
        DIOPTRA_RESTAPI_ENV is set to 'dev'.

    \b
    DIOPTRA_RESTAPI_DATABASE_URI:
        The connection URI for the production database. Used when
        DIOPTRA_RESTAPI_ENV is set to 'prod'.
    """
    pass


@cli.command()
def autoupgrade() -> None:
    """Automatically upgrade the Dioptra database schema to the latest version."""
    if not is_db_uri_set():
        return None

    autoupgrade_cmd(alembic_dir=find_alembic_dir())


@cli.command()
@click.option(
    "--revision",
    type=click.STRING,
    default="head",
    show_default=True,
    help="Revision identifier",
)
def upgrade(revision: str):
    """Upgrade the Dioptra database schema to the specified revision."""
    if not is_db_uri_set():
        return None

    _flask_db_upgrade(alembic_dir=find_alembic_dir(), revision=revision)


@cli.command()
@click.argument(
    "revision",
    nargs=1,
    type=click.STRING,
)
def downgrade(revision: str):
    """Downgrade the Dioptra database schema to an earlier version.

    REVISION: Revision identifier
    """
    if not is_db_uri_set():
        return None

    _flask_db_downgrade(alembic_dir=find_alembic_dir(), revision=revision)


@cli.command()
@click.option(
    "-i",
    "--indicate-current",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Indicate the current revision",
)
def history(indicate_current: bool):
    """Display the Dioptra database schema history."""
    if not is_db_uri_set():
        return None

    _flask_db_history(alembic_dir=find_alembic_dir(), indicate_current=indicate_current)


@cli.command()
@click.option(
    "-d",
    "--directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path("src", "dioptra", "restapi", "db", "alembic"),
    help='Migration script directory (default is "src/dioptra/restapi/db/alembic")',
)
@click.option("-m", "--message", type=click.STRING, help="Revision message")
@click.option(
    "--sql",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--head",
    type=click.STRING,
    help="Specify head revision or <branchname>@head to base new revision on",
)
@click.option(
    "--splice",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help='Allow a non-head revision as the "head" to splice onto',
)
@click.option(
    "--branch-label",
    type=click.STRING,
    help="Specify a branch label to apply to the new revision",
)
@click.option(
    "--version-path",
    type=click.STRING,
    help="Specify specific path from config for version file",
)
@click.option(
    "--rev-id",
    type=click.STRING,
    help="Specify a hardcoded revision id instead of generating one",
)
@click.option(
    "-x",
    "--x-arg",
    type=click.STRING,
    multiple=True,
    help="Additional arguments consumed by custom env.py scripts",
)
def migrate(
    directory: Path,
    message: str | None,
    sql: bool,
    head: str | None,
    splice: bool,
    branch_label: str | None,
    version_path: str | None,
    rev_id: str | None,
    x_arg: list[str],
):
    """Autogenerate a new revision file (Alias for 'flask db migrate')."""
    if not is_db_uri_set():
        return None

    _flask_db_migrate(
        alembic_dir=directory,
        message=message,
        sql=sql,
        head=head,
        splice=splice,
        branch_label=branch_label,
        version_path=version_path,
        rev_id=rev_id,
        x_arg=x_arg,
    )


@cli.command()
@click.option(
    "-d",
    "--directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    default=Path("src", "dioptra", "restapi", "db", "alembic"),
    help='Migration script directory (default is "src/dioptra/restapi/db/alembic")',
)
@click.option("-m", "--message", type=click.STRING, help="Revision message")
@click.option(
    "--autogenerate",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help=(
        "Populate revision script with candidate migration operations, based on "
        "comparison of database to model"
    ),
)
@click.option(
    "--sql",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--head",
    type=click.STRING,
    help="Specify head revision or <branchname>@head to base new revision on",
)
@click.option(
    "--splice",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help='Allow a non-head revision as the "head" to splice onto',
)
@click.option(
    "--branch-label",
    type=click.STRING,
    help="Specify a branch label to apply to the new revision",
)
@click.option(
    "--version-path",
    type=click.STRING,
    help="Specify specific path from config for version file",
)
@click.option(
    "--rev-id",
    type=click.STRING,
    help="Specify a hardcoded revision id instead of generating one",
)
def revision(
    directory: Path,
    message: str | None,
    autogenerate: bool,
    sql: bool,
    head: str | None,
    splice: bool,
    branch_label: str | None,
    version_path: str | None,
    rev_id: str | None,
):
    """Create a new revision file (Alias for 'flask db revision')."""
    if not is_db_uri_set():
        return None

    _flask_db_revision(
        alembic_dir=directory,
        message=message,
        autogenerate=autogenerate,
        sql=sql,
        head=head,
        splice=splice,
        branch_label=branch_label,
        version_path=version_path,
        rev_id=rev_id,
    )


if __name__ == "__main__":
    cli()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform group endpoint operations."""
from __future__ import annotations

import datetime
from typing import Any, Final, cast

import structlog
from injector import inject
from sqlalchemy import func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import group_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
)
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

DEFAULT_GROUP_MEMBER_PERMISSIONS: Final[dict[str, bool]] = {
    "read": False,
    "write": False,
    "share_read": False,
    "share_write": False,
}
DEFAULT_GROUP_MANAGER_PERMISSIONS: Final[dict[str, bool]] = {
    "owner": False,
    "admin": True,
}
GROUP_CREATOR_MEMBER_PERMISSIONS: Final[dict[str, bool]] = {
    "read": True,
    "write": True,
    "share_read": True,
    "share_write": True,
}
GROUP_CREATOR_MANAGER_PERMISSIONS: Final[dict[str, bool]] = {
    "owner": True,
    "admin": True,
}
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.Group.name.like(x, escape="/"),
}

GROUP_TYPE: Final[str] = "group"


class GroupService(object):
    """The service methods used for creating and managing groups."""

    @inject
    def __init__(
        self,
        group_name_service: GroupNameService,
        group_member_service: GroupMemberService,
        group_manager_service: GroupManagerService,
    ) -> None:
        """Initialize the group service.

        All arguments are provided via dependency injection.

        Args:
            group_name_service: A GroupNameService object.
            group_member_service: A GroupMemberService object.
            group_manager_service: A GroupManagerService object.
        """
        self._group_name_service = group_name_service
        self._group_member_service = group_member_service
        self._group_manager_service = group_manager_service

    def create(
        self,
        name: str,
        creator: models.User,
        commit: bool = True,
        **kwargs,
    ) -> models.Group:
        """Create a new group.

        Args:
            name: The name requested by the user.
            creator: The user who created the group.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The new group object.

        Raises:
            GroupNameNotAvailable: If the group name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._group_name_service.get(name)
        if duplicate is not None:
            raise EntityExistsError("group", duplicate.group_id, name=name)

        new_group = models.Group(name=name, creator=creator)
        self._group_member_service.create(
            group=new_group,
            user=creator,
            permissions=GROUP_CREATOR_MEMBER_PERMISSIONS,
            commit=False,
            log=log,
        )
        self._group_manager_service.create(
            group=new_group,
            user=creator,
            permissions=GROUP_CREATOR_MANAGER_PERMISSIONS,
            commit=False,
            log=log,
        )

        db.session.add(new_group)

        if commit:
            db.session.commit()
            log.debug("Group creation successful", group_id=new_group.group_id)

        return new_group

    def get(
        self,
        search_string: str,
        page_index: int,
        page_length: int,
        **kwargs,
    ) -> tuple[list[models.Group], int]:
        """Fetch a list of groups, optionally filtering by search string and paging
        parameters.

        Args:
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of groups to be returned.

        Returns:
            A tuple containing a list of groups and the total number of groups matching
            the query.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of groups.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get list of groups")

        search_filters = construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)

        stmt = (
            select(func.count(models.Group.group_id))
            .filter_by(is_deleted=False)
            .filter(search_filters)
        )
        total_num_groups = db.session.scalars(stmt).first()

        if total_num_groups is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_groups == 0:
            return cast(list[models.Group], []), total_num_groups

        stmt = (
            select(models.Group)  # type: ignore
            .filter_by(is_deleted=False)
            .filter(search_filters)
            .offset(page_index)
            .limit(page_length)
        )
        groups = cast(list[models.Group], db.session.scalars(stmt).all())

        return groups, total_num_groups


class GroupIdService(object):
    """The service methods used to manage a group by ID."""

    @inject
    def __init__(
        self, group_service: GroupService, group_name_service: GroupNameService
    ) -> None:
        """Initialize the group ID service.

        All arguments are provided via dependency injection.

        Args:
            group_service: A GroupService object.
            group_name_service: A GroupNameService object.
        """
        self._group_service = group_service
        self._group_name_service = group_name_service

    def get(
        self, group_id: int, error_if_not_found: bool = False, **kwargs
    ) -> models.Group | None:
        """Fetch a group by its unique id.

        Args:
            group_id: The unique id of the group.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.

        Returns:
            The group object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the group is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Lookup group by unique id", group_id=group_id)

        stmt = select(models.Group).filter_by(group_id=group_id, is_deleted=False)
        group: models.Group | None = db.session.scalars(stmt).first()

        if group is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(GROUP_TYPE, group_id=group_id)

            return None

        return group

    def modify(
        self,
        group_id: int,
        name: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> models.Group | None:
        """Modifies the specified group.

        Args:
            group_id: The ID of the group to be modified.
            name: The new name for the group.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The group object.

        Raises:
            EntityDoesNotExistError: If the group is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Modify group", group_id=group_id)

        stmt = select(models.Group).filter_by(group_id=group_id, is_deleted=False)
        group = db.session.scalars(stmt).first()

        if group is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(GROUP_TYPE, group_id=group_id)

            return None

        duplicate = self._group_name_service.get(name, log=log)
        if duplicate is not None:
            raise EntityExistsError(GROUP_TYPE, duplicate.group_id, name=name)

        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        group.last_modified_on = current_timestamp
        group.name = name

        if commit:
            db.session.commit()

        return group

    def delete(self, group_id: int, **kwargs) -> dict[str, Any]:
        """Permanently deletes the group by ID.

        Args:
            group_id: The ID of the group to be deleted.

        Returns:
            A dictionary containing the delete group success message if the group is
            deleted successfully.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Delete group", group_id=group_id)

        stmt = select(models.Group).filter_by(group_id=group_id, is_deleted=False)
        group = db.session.scalars(stmt).first()

        if group is None:
            raise EntityDoesNotExistError(GROUP_TYPE, group_id=group_id)

        name = group.name

        deleted_group_lock = models.GroupLock(
            group_lock_type=group_lock_types.DELETE,
            group=group,
        )
        db.session.add(deleted_group_lock)
        db.session.commit()
        log.debug("Group deleted", group_id=group.group_id)

        return {"status": "Success", "group": name}


class GroupNameService(object):
    """The service methods used to manage a group by name."""

    def get(
        self, name: str, error_if_not_found: bool = False, **kwargs
    ) -> models.Group | None:
        """Fetch a group by its name.

        Args:
            name: The name of the group.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.

        Returns:
            The group object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the group is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Lookup group by name", name=name)

        stmt = select(models.Group).filter_by(name=name)
        group = db.session.scalars(stmt).first()

        if group is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(GROUP_TYPE, name=name)

            return None

        return group


class GroupMemberService(object):
    """The service methods used to manage a group's members."""

    def create(
        self,
        group: models.Group,
        user: models.User,
        permissions: dict[str, bool] | None = None,
        commit: bool = True,
        **kwargs,
    ) -> models.GroupMember:
        """Add a user to a group.

        If not specified, the default permissions will be::

            {
                "read": False,
                "write": False,
                "share_read": False,
                "share_write": False,
            }

        Args:
            group: The group to which the user is being added.
            user: The user being added to the group.
            permissions: The permissions for the user in the group. Defaults to None.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The new group member object.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        permissions = permissions or DEFAULT_GROUP_MEMBER_PERMISSIONS
        log.debug("Add group member", group_id=group.group_id, user_id=user.user_id)
        group_member = models.GroupMember(
            user=user,
            read=permissions["read"],
            write=permissions["write"],
            share_read=permissions["share_read"],
            share_write=permissions["share_write"],
        )
        group.members.append(group_member)

        if commit:
            db.session.commit()

        return group_member


class GroupManagerService(object):
    """The service methods used to manage a group's managers."""

    def create(
        self,
        group: models.Group,
        user: models.User,
        permissions: dict[str, bool] | None = None,
        commit: bool = True,
        **kwargs,
    ) -> models.GroupManager:
        """Assign an existing group member to be a manager of a group.

        If not specified, the default permissions will be::

            {
                "owner": False,
                "admin": True,
            }

        Args:
            group: The group receiving a new manager.
            user: The user being assigned as a manager of the group. The user must
                already be a member of the group.
            permissions: The permissions for the user in the group. Defaults to None.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The new group manager object.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        permissions = permissions or DEFAULT_GROUP_MANAGER_PERMISSIONS
        log.debug("Add group manager", group_id=group.group_id, user_id=user.user_id)
        group_manager = models.GroupManager(
            user=user,
            owner=permissions["owner"],
            admin=permissions["admin"],
        )
        group.managers.append(group_manager)

        if commit:
            db.session.commit()

        return group_manager



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Tuple

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.bounding_boxes.coordinates import (
    TensorflowBoundingBoxesBatchedGrid,
)

from .bounding_boxes_iou import BoundingBoxesBatchedGridIOU, BoundingBoxesIOU

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowBoundingBoxesIOU(BoundingBoxesIOU):
    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def iou(self, bbox_corner1, bbox_corner2):
        x1 = tf.maximum(bbox_corner1[..., 0], bbox_corner2[..., 0])
        y1 = tf.maximum(bbox_corner1[..., 1], bbox_corner2[..., 1])
        x2 = tf.minimum(bbox_corner1[..., 2], bbox_corner2[..., 2])
        y2 = tf.minimum(bbox_corner1[..., 3], bbox_corner2[..., 3])

        intersection_area = tf.maximum(0.0, (x2 - x1)) * tf.maximum(0.0, (y2 - y1))
        box_1_area = (bbox_corner1[..., 2] - bbox_corner1[..., 0]) * (
            bbox_corner1[..., 3] - bbox_corner1[..., 1]
        )
        box_2_area = (bbox_corner2[..., 2] - bbox_corner2[..., 0]) * (
            bbox_corner2[..., 3] - bbox_corner2[..., 1]
        )
        union_area = box_1_area + box_2_area - intersection_area

        return intersection_area / union_area


class TensorflowBoundingBoxesBatchedGridIOU(BoundingBoxesBatchedGridIOU):
    def __init__(
        self,
        bounding_boxes_iou: TensorflowBoundingBoxesIOU,
        bounding_boxes_batched_grid: TensorflowBoundingBoxesBatchedGrid,
    ) -> None:
        self._bounding_boxes_iou = bounding_boxes_iou
        self._bbox_batched_grid = bounding_boxes_batched_grid

    @classmethod
    def on_grid_shape(
        cls, grid_shape: Tuple[int, int]
    ) -> TensorflowBoundingBoxesBatchedGridIOU:
        return cls(
            bounding_boxes_iou=TensorflowBoundingBoxesIOU(),
            bounding_boxes_batched_grid=(
                TensorflowBoundingBoxesBatchedGrid.on_grid_shape(grid_shape=grid_shape)
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def iou(self, bboxes_cell_xywh1: Tensor, bboxes_cell_xywh2: Tensor) -> Tensor:
        n_bounding_boxes1 = tf.cast(tf.shape(bboxes_cell_xywh1)[-2], tf.int32)
        n_bounding_boxes2 = tf.cast(tf.shape(bboxes_cell_xywh2)[-2], tf.int32)
        max_n_bounding_boxes = tf.cast(n_bounding_boxes1 * n_bounding_boxes2, tf.int32)

        bboxes_corner1 = self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh1, n_bounding_boxes=n_bounding_boxes1
        )
        bboxes_corner2 = self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh2, n_bounding_boxes=n_bounding_boxes2
        )

        tiled_bboxes_corner1 = tf.cond(
            pred=tf.math.greater(max_n_bounding_boxes, n_bounding_boxes1),
            true_fn=lambda: tf.tile(
                bboxes_corner1, multiples=(1, 1, 1, n_bounding_boxes2, 1)
            ),
            false_fn=lambda: bboxes_corner1,
        )
        tiled_bboxes_corner2 = tf.cond(
            pred=tf.math.greater(max_n_bounding_boxes, n_bounding_boxes2),
            true_fn=lambda: tf.tile(
                bboxes_corner2, multiples=(1, 1, 1, n_bounding_boxes1, 1)
            ),
            false_fn=lambda: bboxes_corner2,
        )

        return self._bounding_boxes_iou.iou(tiled_bboxes_corner1, tiled_bboxes_corner2)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def max_iou(self, bboxes_cell_xywh1: Tensor, bboxes_cell_xywh2: Tensor) -> Tensor:
        iou_areas = self.iou(bboxes_cell_xywh1, bboxes_cell_xywh2)

        return tf.reduce_max(iou_areas, axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def select_max_iou_bboxes(
        self,
        bboxes_cell_xywh: Tensor,
        bboxes_conf: Tensor,
        bboxes_cell_xywh_ground_truth: Tensor,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        batch_size = tf.cast(tf.shape(bboxes_cell_xywh)[0], tf.int32)
        n_bounding_boxes = tf.cast(tf.shape(bboxes_cell_xywh)[-2], tf.int32)
        bboxes_corner = self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh, n_bounding_boxes=n_bounding_boxes
        )
        bboxes_corner_ground_truth = self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh_ground_truth, n_bounding_boxes=1
        )

        iou_areas = self.iou(bboxes_corner, bboxes_corner_ground_truth)

        batch_indices = self._generate_flattened_batch_indices(batch_size=batch_size)
        dim_i_indices = self._generate_flattened_dim_i_indices(batch_size=batch_size)
        dim_j_indices = self._generate_flattened_dim_j_indices(batch_size=batch_size)

        sorted_iou_indices = tf.math.top_k(iou_areas, k=n_bounding_boxes).indices
        max_iou_indices = tf.reshape(
            sorted_iou_indices[..., :1],
            (
                batch_size,
                self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol,
                1,
            ),
        )
        non_max_iou_indices = tf.reshape(
            sorted_iou_indices[..., 1:],
            (
                batch_size,
                self._bbox_batched_grid.cell_nrow
                * self._bbox_batched_grid.cell_ncol
                * (n_bounding_boxes - 1),
                1,
            ),
        )

        bboxes_gather_indices = self._prepare_gather_indices(
            batch_size=batch_size,
            n_bounding_boxes=1,
            batch_indices=batch_indices,
            dim_i_indices=dim_i_indices,
            dim_j_indices=dim_j_indices,
            bbox_indices=max_iou_indices,
        )
        no_bboxes_gather_indices = self._prepare_gather_indices(
            batch_size=batch_size,
            n_bounding_boxes=n_bounding_boxes - 1,
            batch_indices=batch_indices,
            dim_i_indices=dim_i_indices,
            dim_j_indices=dim_j_indices,
            bbox_indices=non_max_iou_indices,
        )

        selected_bboxes = tf.reshape(
            tf.gather_nd(params=bboxes_cell_xywh, indices=bboxes_gather_indices),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow,
                self._bbox_batched_grid.cell_ncol,
                4,
            ),
        )
        selected_bboxes_conf = tf.reshape(
            tf.gather_nd(params=bboxes_conf, indices=bboxes_gather_indices),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow,
                self._bbox_batched_grid.cell_ncol,
            ),
        )
        selected_no_bboxes_conf = tf.reshape(
            tf.gather_nd(params=bboxes_conf, indices=no_bboxes_gather_indices),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow,
                self._bbox_batched_grid.cell_ncol,
                n_bounding_boxes - 1,
            ),
        )

        return selected_bboxes, selected_bboxes_conf, selected_no_bboxes_conf

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def _generate_flattened_batch_indices(self, batch_size: Tensor) -> Tensor:
        batch_indices = tf.repeat(
            tf.range(batch_size),
            repeats=self._bbox_batched_grid.cell_nrow
            * self._bbox_batched_grid.cell_ncol,
        )
        batch_indices = tf.reshape(
            batch_indices,
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol,
                1,
            ),
        )

        return tf.cast(batch_indices, tf.int32)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def _generate_flattened_dim_i_indices(self, batch_size: Tensor) -> Tensor:
        dim_i_indices = tf.transpose(
            tf.tile(
                tf.expand_dims(
                    tf.range(self._bbox_batched_grid.cell_nrow, dtype=tf.int32), axis=0
                ),
                multiples=[self._bbox_batched_grid.cell_ncol, 1],
            )
        )
        dim_i_indices = tf.reshape(
            dim_i_indices,
            (self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol, 1),
        )
        dim_i_indices = tf.broadcast_to(
            dim_i_indices,
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol,
                1,
            ),
        )

        return tf.cast(dim_i_indices, tf.int32)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def _generate_flattened_dim_j_indices(self, batch_size: Tensor) -> Tensor:
        dim_j_indices = tf.tile(
            tf.expand_dims(
                tf.range(self._bbox_batched_grid.cell_ncol, dtype=tf.int32), axis=0
            ),
            multiples=[self._bbox_batched_grid.cell_nrow, 1],
        )
        dim_j_indices = tf.reshape(
            dim_j_indices,
            (self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol, 1),
        )
        dim_j_indices = tf.broadcast_to(
            dim_j_indices,
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow * self._bbox_batched_grid.cell_ncol,
                1,
            ),
        )

        return tf.cast(dim_j_indices, tf.int32)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def _prepare_gather_indices(
        self,
        batch_size: Tensor,
        n_bounding_boxes: Tensor,
        batch_indices: Tensor,
        dim_i_indices: Tensor,
        dim_j_indices: Tensor,
        bbox_indices: Tensor,
    ) -> Tensor:
        batch_indices = tf.reshape(
            tf.tile(batch_indices, multiples=(1, 1, n_bounding_boxes)),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow
                * self._bbox_batched_grid.cell_ncol
                * n_bounding_boxes,
                1,
            ),
        )
        dim_i_indices = tf.reshape(
            tf.tile(dim_i_indices, multiples=(1, 1, n_bounding_boxes)),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow
                * self._bbox_batched_grid.cell_ncol
                * n_bounding_boxes,
                1,
            ),
        )
        dim_j_indices = tf.reshape(
            tf.tile(dim_j_indices, multiples=(1, 1, n_bounding_boxes)),
            shape=(
                batch_size,
                self._bbox_batched_grid.cell_nrow
                * self._bbox_batched_grid.cell_ncol
                * n_bounding_boxes,
                1,
            ),
        )

        return tf.concat(
            [batch_indices, dim_i_indices, dim_j_indices, bbox_indices], axis=2
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from abc import ABCMeta, abstractmethod


class BoundingBoxCoordinates(metaclass=ABCMeta):
    @property
    @abstractmethod
    def cell_height(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_width(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_ncol(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_nrow(self):
        raise NotImplementedError

    @abstractmethod
    def apply_constraint_one_object_per_cell(
        self, bboxes_cell_xywh, bboxes_cell_ij, bboxes_labels
    ):
        raise NotImplementedError

    @abstractmethod
    def from_corner_to_image_xywh(self, bboxes_corner):
        raise NotImplementedError

    @abstractmethod
    def from_image_to_cell_xywh(self, bboxes_image_xywh):
        raise NotImplementedError

    @abstractmethod
    def from_corner_to_cell_xywh(self, bboxes_corner):
        raise NotImplementedError

    @abstractmethod
    def from_cell_xywh_to_image_xywh(self, bboxes_cell_xywh, bboxes_cell_ij):
        raise NotImplementedError

    @abstractmethod
    def from_image_xywh_to_corner(self, bboxes_image_xywh):
        raise NotImplementedError

    @abstractmethod
    def from_cell_xywh_to_corner(self, bboxes_cell_xywh, bboxes_cell_ij):
        raise NotImplementedError


class BoundingBoxesBatchedGrid(metaclass=ABCMeta):
    @property
    @abstractmethod
    def cell_height(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_width(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_ncol(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def cell_nrow(self):
        raise NotImplementedError

    @abstractmethod
    def embed(self, bboxes_corner, bboxes_labels, n_classes):
        raise NotImplementedError

    @abstractmethod
    def extract_using_mask(self, bboxes_grid, labels_grid, cell_mask):
        raise NotImplementedError

    @abstractmethod
    def from_corner_to_image_xywh(self, bboxes_corner):
        raise NotImplementedError

    @abstractmethod
    def from_image_to_cell_xywh(self, bboxes_image_xywh):
        raise NotImplementedError

    @abstractmethod
    def from_corner_to_cell_xywh(self, bboxes_corner):
        raise NotImplementedError

    @abstractmethod
    def from_cell_xywh_to_image_xywh(self, bboxes_cell_xywh, n_bounding_boxes):
        raise NotImplementedError

    @abstractmethod
    def from_image_xywh_to_corner(self, bboxes_image_xywh):
        raise NotImplementedError

    @abstractmethod
    def from_cell_xywh_to_corner(self, bboxes_cell_xywh, n_bounding_boxes):
        raise NotImplementedError



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A module for registering the endpoint routes with the main application.

.. |Api| replace:: :py:class:`flask_restx.Api`
"""
from __future__ import annotations

from flask_restx import Api

V1_ROOT = "api/v1"

V1_ARTIFACTS_ROUTE = "artifacts"
V1_AUTH_ROUTE = "auth"
V1_ENTRYPOINTS_ROUTE = "entrypoints"
V1_EXPERIMENTS_ROUTE = "experiments"
V1_GROUPS_ROUTE = "groups"
V1_JOBS_ROUTE = "jobs"
V1_MODELS_ROUTE = "models"
V1_PLUGIN_FILES_ROUTE = "files"
V1_PLUGIN_PARAMETER_TYPES_ROUTE = "pluginParameterTypes"
V1_PLUGINS_ROUTE = "plugins"
V1_QUEUES_ROUTE = "queues"
V1_TAGS_ROUTE = "tags"
V1_USERS_ROUTE = "users"
V1_WORKFLOWS_ROUTE = "workflows"


def register_routes(api: Api) -> None:
    """Registers the endpoint routes with the main application.

    Args:
        api: The main REST |Api| object.
    """
    from .v1.artifacts.controller import api as artifacts_api
    from .v1.auth.controller import api as auth_api
    from .v1.entrypoints.controller import api as entrypoints_api
    from .v1.experiments.controller import api as experiments_api
    from .v1.groups.controller import api as groups_api
    from .v1.jobs.controller import api as jobs_api
    from .v1.models.controller import api as models_api
    from .v1.plugin_parameter_types.controller import api as plugin_parameter_types_api
    from .v1.plugins.controller import api as plugins_api
    from .v1.queues.controller import api as queues_api
    from .v1.tags.controller import api as tags_api
    from .v1.users.controller import api as users_api
    from .v1.workflows.controller import api as workflows_api

    api.add_namespace(auth_api, path=f"/{V1_ROOT}/{V1_AUTH_ROUTE}")
    api.add_namespace(artifacts_api, path=f"/{V1_ROOT}/{V1_ARTIFACTS_ROUTE}")
    api.add_namespace(entrypoints_api, path=f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}")
    api.add_namespace(experiments_api, path=f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}")
    api.add_namespace(groups_api, path=f"/{V1_ROOT}/{V1_GROUPS_ROUTE}")
    api.add_namespace(jobs_api, path=f"/{V1_ROOT}/{V1_JOBS_ROUTE}")
    api.add_namespace(models_api, path=f"/{V1_ROOT}/{V1_MODELS_ROUTE}")
    api.add_namespace(
        plugin_parameter_types_api, path=f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}"
    )
    api.add_namespace(plugins_api, path=f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}")
    api.add_namespace(queues_api, path=f"/{V1_ROOT}/{V1_QUEUES_ROUTE}")
    api.add_namespace(tags_api, path=f"/{V1_ROOT}/{V1_TAGS_ROUTE}")
    api.add_namespace(users_api, path=f"/{V1_ROOT}/{V1_USERS_ROUTE}")
    api.add_namespace(workflows_api, path=f"/{V1_ROOT}/{V1_WORKFLOWS_ROUTE}")



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Queue resources."""
from marshmallow import Schema, fields

from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)

QueueRefBaseSchema = generate_base_resource_ref_schema("Queue")
QueueSnapshotRefBaseSchema = generate_base_resource_ref_schema(
    "Queue", keep_snapshot_id=True
)


class QueueRefSchema(QueueRefBaseSchema):  # type: ignore
    """The reference schema for the data stored in a Queue resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Queue resource."),
    )


class QueueSnapshotRefSchema(QueueSnapshotRefBaseSchema):  # type: ignore
    """The snapshot reference schema for the data stored in a Queue resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Queue resource."),
    )


class QueueMutableFieldsSchema(Schema):
    """The fields schema for the mutable data in a Queue resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Queue resource."),
        required=True,
    )
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Queue resource."),
        load_default=None,
    )


QueueBaseSchema = generate_base_resource_schema("Queue", snapshot=True)


class QueueSchema(QueueMutableFieldsSchema, QueueBaseSchema):  # type: ignore
    """The schema for the data stored in a Queue resource."""


class QueuePageSchema(BasePageSchema):
    """The paged schema for the data stored in a Queue resource."""

    data = fields.Nested(
        QueueSchema,
        many=True,
        metadata=dict(description="List of Queue resources in the current page."),
    )


class QueueGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /queues endpoint."""
